@@ -5,16 +5,16 @@
 #==============================================================================
 
 import networkx as nx
-import pickle
 import json
 import numpy as np
 import random
 import sys
+import warnings
 
 from rdkit import Chem
 from rdkit.Chem import Conformer, rdMolAlign
 from rdkit.Geometry import Point3D
-from collections import Counter, defaultdict
+from collections import Counter, defaultdict, OrderedDict
 from itertools import combinations, product
 from hashlib import md5
 from xml.dom import minidom
@@ -49,12 +49,18 @@ class __PharmacophoreBase():
         self.__cached = cached
         self.__cache = dict()
 
+    def __drop_cache(self):
+        self.__cache = dict()
+
     @staticmethod
     def __remove_dupl(ls):
         # remove duplicates preserving the order of items
         seen = set()
         seen_add = seen.add
-        return [x for x in ls if not (x in seen or seen_add(x))]
+        output = [x for x in ls if not (x in seen or seen_add(x))]
+        if len(output) < len(ls):
+            warnings.warn('Features with identical labels and coordinates were found and duplicates were removed')
+        return output
 
     def load_from_feature_coords(self, feature_coords):
         # feature_coords: [('A', (1.23, 2.34, 3.45)), ('A', (4.56, 5.67, 6.78)), ...]
@@ -65,6 +71,22 @@ class __PharmacophoreBase():
             self.__g.add_node(i, label=label, xyz=coords)
         self.__update_dists()
 
+    def add_feature(self, label, xyz):
+        """
+        Add a feature with its coordinates to the pharmacophore
+        :param label: text label of the feature
+        :param xyz: tuple of xyz coordinates of the feature
+        :return: id of the added feature
+        """
+        if self.__g.nodes:
+            feature_id = max(list(self.__g.nodes)) + 1
+        else:
+            feature_id = 0
+        self.__g.add_node(feature_id, label=label, xyz=tuple(xyz))
+        self.__drop_cache()
+        self.__update_dists()
+        return feature_id
+
     def __update_dists(self, bin_step=None):
         for i, j in combinations(self.__g.nodes(), 2):
             dist = self.__dist(self.__g.nodes[i]['xyz'], self.__g.nodes[j]['xyz'], bin_step)
@@ -112,9 +134,13 @@ class __PharmacophoreBase():
             d[res] += 1
         return d
 
-    def __get_full_hash(self, ids=None, tol=0):
+    def __get_full_hash(self, ids=None, tol=0, hex=True):
         d = self.__get_signature_dict(ids, tol)
-        return md5(pickle.dumps(str(tuple(sorted(d.items()))))).hexdigest()
+        h = md5(str(tuple(sorted(d.items()))).encode())
+        if hex:
+            return h.hexdigest()
+        else:
+            return h.digest()
 
     def __gen_quadruplet_canon_name_stereo(self, feature_ids, tol=0):
         # return canon quadruplet signature and stereo
@@ -261,7 +287,7 @@ class __PharmacophoreBase():
         data = self.__g.nodes(data=True)
         return Counter([item[1]['label'] for item in data])
 
-    def get_signature_md5(self, ids=None, tol=0):
+    def get_signature_md5(self, ids=None, tol=0, hex=True):
         """
         Returns pharmacophore hash.
 
@@ -271,11 +297,13 @@ class __PharmacophoreBase():
                     between an edge and a plane formed by other three features. Quadruplets having at least one angle
                     less than tolerance value are assigned 0 chirality.
         :type tol: float
+        :param hex: return hex representation (True) or binary (False)
+        :type hex: bool
         :return: md5 hash of a pharmacophore
         :rtype: str
 
         """
-        return self.__get_full_hash(self._get_ids(ids), tol)
+        return self.__get_full_hash(self._get_ids(ids), tol, hex)
 
     def get_feature_coords(self, ids=None):
         """
@@ -291,6 +319,17 @@ class __PharmacophoreBase():
         else:
             return [(v['label'], v['xyz']) for k, v in self.__g.nodes(data=True) if k in set(ids)]
 
+    def get_feature_ids(self):
+        """
+        Returns a dict of feature labels and a tuple of corresponding features ids
+
+        :return: dist {label: (id1, id2, ...), ...}
+        """
+        output = defaultdict(list)
+        for i, data in self.__g.nodes(data=True):
+            output[data['label']].append(i)
+        return {k: tuple(v) for k, v in output.items()}
+
     def get_mirror_pharmacophore(self):
         """
         Returns a new mirrored Pharmacophore instance.
@@ -426,7 +465,7 @@ class __PharmacophoreBase():
                         output[(nbits_, act_bits_, tol_)].add(random.randrange(nbits_))
         return output
 
-    def get_descriptors(self, tol=0, ncomb=4):
+    def get_descriptors(self, tol=0, ncomb=4, ids=None):
         """
         Returns count-based descriptor string of a pharmacophore.
 
@@ -434,6 +473,8 @@ class __PharmacophoreBase():
         :type tol: float
         :param ncomb: number of feature combinations in descriptors, can be an integer from 1 to 4
         :type ncomb: int
+        :param ids: feature ids used to calculate descriptors
+        :type ids: list
         :return: dictionary where keys are signatures of quadruplets and values are counts of quadruples with
                  identical signatures.
                  Signature of a quadruplet is for example 'AA2A2D2|AA2A3D0|AA2A3D3|DA0A2A3|0|0|1';
@@ -444,7 +485,7 @@ class __PharmacophoreBase():
         """
         if not isinstance(ncomb, int) or ncomb < 1 or ncomb > 4:
             return tuple()
-        ids = self._get_ids(None)
+        ids = self._get_ids(ids)
         d = self.__get_signature_dict(ids, tol, ncomb)
         return {k[2:-1].replace("', ", '|').replace(", ", '|') + f'|{self.__bin_step}': v for k, v in d.items()}
 
@@ -456,7 +497,7 @@ class __PharmacophoreMol(__PharmacophoreBase):
 
     """
 
-    __feat_dict_mol = {'A': 89, 'P': 15, 'N': 7, 'H': 1, 'D': 66, 'a': 10}
+    __feat_dict_mol = {'A': 89, 'P': 15, 'N': 7, 'H': 1, 'D': 66, 'a': 10, 'T': 11}
 
     def __init__(self, bin_step=1, cached=False):
         super().__init__(bin_step, cached)
@@ -576,14 +617,17 @@ class __PharmacophoreMatch(__PharmacophoreMol):
         return None
 
 
-class Pharmacophore(__PharmacophoreMatch):
+class __PharmacophoreLoadedMol(__PharmacophoreMatch):
 
     """
-    Main class
-
+    This class manages a loaded molecule
     """
 
-    __feat_dict_ls = {"A": "HBA", "H": "H", "D": "HBD", "P": "PI", "N": "NI", "a": "AR"}
+    def __init__(self, bin_step=1, cached=False):
+        super().__init__(bin_step, cached)
+        self.atom_ids = None    # will store list of tuples (('A', (1,)), ('P', (5,6,7)), ...),
+                                # the order should be preserved while reading to keep the correspondence with
+                                # feature ids which will be computed in load_from_feature_coords
 
     def load_from_mol(self, mol):
         """
@@ -637,26 +681,24 @@ class Pharmacophore(__PharmacophoreMatch):
                 res.append(tuple(item))
         return tuple(res)
 
-    @staticmethod
-    def _get_features_atom_ids_factory(mol, factory):
+    def _get_features_atom_ids_factory(self, mol, factory):
         """
         function works with RDKit feature factory
         output is dict
         {'N': ((1,), (4,5,6)), 'P': ((2,)), 'a': ((7,8,9,10,11)), ...}
         """
         features = factory.GetFeaturesForMol(Chem.AddHs(mol))
-        output = dict()
+        output = OrderedDict()
         for f in features:
             name = f.GetFamily()
             if name in output:
                 output[name].append(f.GetAtomIds())
             else:
                 output[name] = [f.GetAtomIds()]
-        output = {k: Pharmacophore.__filter_features(v) for k, v in output.items()}
+        output = {k: self.__filter_features(v) for k, v in output.items()}
         return output
 
-    @staticmethod
-    def _get_features_atom_ids(mol, smarts_features):
+    def _get_features_atom_ids(self, mol, smarts_features):
         """
         function works with explicit smarts definitions of features
         output is dict
@@ -672,7 +714,7 @@ class Pharmacophore(__PharmacophoreMatch):
             ids = [tuple(sorted(i)) for i in ids]
             return ids
 
-        output = dict()
+        output = OrderedDict()
 
         m = Chem.AddHs(mol)
 
@@ -684,7 +726,7 @@ class Pharmacophore(__PharmacophoreMatch):
                     tmp.extend(ids)
             # exclude features if their atom ids is full subset of another feature of the same type
             if tmp:
-                res = Pharmacophore.__filter_features(tmp)
+                res = self.__filter_features(tmp)
                 output[name] = tuple(res)  # tuple of tuples with atom ids
             else:
                 output[name] = tuple(tmp)  # empty set of ids
@@ -721,13 +763,54 @@ class Pharmacophore(__PharmacophoreMatch):
                     z += coord.z
                 return x / len(ids), y / len(ids), z / len(ids)
 
+        self.atom_ids = []
         feature_coords = []
         for name, features in atom_features_ids.items():
             for feature in features:
                 feature_coords.append((name, get_feature_coord(mol, feature)))
+                self.atom_ids.append((name, feature))
         self.load_from_feature_coords(feature_coords)
+        self.atom_ids = tuple(self.atom_ids)
         return 0
 
+    def get_descriptors_atom_exclusion(self, natoms, tol=0, ncomb=4):
+        """
+        Return descriptors for subsets of features excluding every atom in an input molecule
+
+        :param natoms: total number of heavy atoms in a molecule
+        :type natoms: int
+        :param tol: tolerance
+        :type tol: float
+        :param ncomb: number of feature combinations in descriptors, can be an integer from 1 to 4
+        :type ncomb: int
+        :return: list of dictionaries where keys are signatures of quadruplets and values are counts of quadruples with
+                 identical signatures.
+                 Signature of a quadruplet is for example 'AA2A2D2|AA2A3D0|AA2A3D3|DA0A2A3|0|0|1';
+                 first blocks are signatures of separate features, then configuration sign, tolerance and
+                 binning step value.
+                 Every entry of the list is the descriptors calculated with removal of an atom of
+                 the corresponding index, the first entry - the first atom was removed.
+        :rtype: dict
+        """
+        output = []
+        for i in range(natoms):
+            ids = []
+            for j, item in enumerate(self.atom_ids):
+                if i not in item[1]:
+                    ids.append(j)
+            output.append(self.get_descriptors(tol=tol, ncomb=ncomb, ids=ids))
+        return output
+
+
+class Pharmacophore(__PharmacophoreLoadedMol):
+
+    """
+    Main class
+
+    """
+
+    __feat_dict_ls = {"A": "HBA", "H": "H", "D": "HBD", "P": "PI", "N": "NI", "a": "AR"}
+
     def load_ls_model(self, pml_fname):
         """
         Reads pharmacophore from LigandScout pml-file
@@ -894,3 +977,49 @@ class Pharmacophore(__PharmacophoreMatch):
         else:
             sys.stderr.write('Unknown extension %s. Pharmacophore was not loaded. '
                              'Only pml, pma and xyz formats are supported.' % fname)
+
+    def load_from_pharmit(self, fname):
+        """
+        Reads pharmacophore from pharmit json-file. Only enabled features will be read. Directed features will be
+        reduced to undirected ones (arrows to spheres). Radius will be ignored at read.
+
+        :param fname: json-file name with a pharmit pharmacophore model
+        :return: nothing
+        """
+        feature_names_dict = {'HydrogenDonor': 'D', 'HydrogenAcceptor': 'A', 'NegativeIon': 'N', 'PositiveIon': 'P',
+                              'Hydrophobic': 'H', 'Aromatic': 'a'}
+
+        coords = []
+        p = json.load(open(fname))['points']
+        for i in p:
+            if i['enabled']:
+                # use pmapper feature labels if they are in the list of pre-defined (standard) features
+                label = feature_names_dict[i['name']] if i['name'] in feature_names_dict.keys() else i['name']
+                coords.append((label, (i['x'], i['y'], i['z'])))
+        self.load_from_feature_coords(tuple(coords))
+
+    def save_to_pharmit(self, fname, feature_ids=None, ndigits=2):
+        """
+        Save pharmacophore in pharmit format. All radius will be set to bin_step of the pharmacophore. Labels absent
+        in the standard feature dictionary will be saved as is.
+
+        :param fname: json-file name
+        :param feature_ids: ids of features which should be stored. Default: None (all features).
+        :param ndigits: number of digits after the dot to save coordinates
+        :return: nothing
+        """
+        feature_names_dict = {'D': 'HydrogenDonor', 'A': 'HydrogenAcceptor', 'N': 'NegativeIon', 'P': 'PositiveIon',
+                              'H': 'Hydrophobic', 'a': 'Aromatic'}
+
+        data = {'points': []}
+        bin_step = self.get_bin_step()
+        for label, (x, y, z) in self.get_feature_coords(ids=feature_ids):
+            label = feature_names_dict[label] if label in feature_names_dict.keys() else label
+            data['points'].append({'name': label,
+                                   'x': round(x, ndigits),
+                                   'y': round(y, ndigits),
+                                   'z': round(z, ndigits),
+                                   'radius': bin_step,
+                                   'enabled': True})
+        with open(fname, 'wt') as f:
+            json.dump(data, f, indent=4)
